/**
 *  EditFunctionDialog.java
 *
 *  Original version by  Sreekanth Nagarajan, rewritten
 *  from scratch by Fabio Cozman.
 *
 */

package JavaBayesInterface;

import InferenceGraphs.*;

import java.awt.*;
import java.util.*;

class EditFunctionDialog extends Dialog {
    InferenceGraph ig;
    InferenceGraphNode node;

    int index_extreme_point;

    Panel gp, dp, pp, ap, csp, okp;
    Label parents_announce, parents_labels[];
    Label node_values_labels[];
    Label first_parent_name_label, first_parent_values_labels[], credal_set_label;
    Label separator_label;
    TextField fields[][];
    Choice parent_choices[], credal_set_choice;
    Button return_button;
    
    private final static int TOP_INSET = 5;
    private final static int LEFT_INSET = 10;
    private final static int RIGHT_INSET = 10;
    private final static int BOTTOM_INSET = 0;

    final static String parents_announce_label = "Values for parents:";
    final static String credal_set = "Index of extreme distribution:";
    final static String return_label = "Close window";

    final static Color color_parents = Color.red;
    
    /* ****************************************************** *
     * Default constructor for an EditFunctionDialog object.  *
     * ****************************************************** */
    public EditFunctionDialog(Frame parent,
    InferenceGraph i_g, InferenceGraphNode node) {
    	super(parent, "Edit: " + node.get_name(), 
              false); // This used to be true, but bug in JDK1.1.6 does not accept it.
    	this.ig = i_g;
     	this.node = node;

     	int parents_size = node.get_parents().size();

     	if (parents_size > 1)
     	    build_with_more_than_one_parent();
     	else if (parents_size == 1)
     	    build_with_one_parent();
     	else if (parents_size == 0)
     	    build_with_no_parent();

        // Set the bottom panel with the return button
        okp = new Panel();
        okp.setLayout(new FlowLayout(FlowLayout.CENTER));
        return_button = new Button(return_label);
        okp.add(return_button);

        // Set the final layout
        setLayout(new BorderLayout());
        add("North", gp);
        add("South", okp);

    	pack();
    }

    /* ************************************************ *
     * Customized show method.                          *
     * ************************************************ */
    public synchronized void show() {
    	Rectangle bounds = getParent().bounds();
    	Rectangle abounds = bounds();

    	move(bounds.x + (bounds.width - abounds.width)/ 2,
    	     bounds.y + (bounds.height - abounds.height)/2); 

    	super.show();
    }
    
    /* ************************************************** *
     * Customize insets method.                           *
     * ************************************************** */
    public Insets insets() {
        Insets ins = super.insets();
        return(new Insets(ins.top + TOP_INSET, ins.left + LEFT_INSET,
                          ins.bottom + BOTTOM_INSET, ins.right + RIGHT_INSET));
    }
    
    /* ******************************************************* *
     * Build the widget when the variable has more than one    *
     * parent.                                                 *
     * ******************************************************* */
    private void build_with_more_than_one_parent() {
        // To avoid repetition, collect some information
        // about the node.
      	Vector node_parents = node.get_parents();
      	String node_values[] = node.get_values();
      	int parents_size = node_parents.size();

        // Information about a generic node that is parent of
        // the given node.
        InferenceGraphNode parent_node;
        String parent_node_values[];

        // Information about the first parent of the given node.
        parent_node = (InferenceGraphNode)(node_parents.firstElement());
        String first_parent_name = parent_node.get_name();
        String first_parent_values[] = parent_node.get_values();

        // Construct the panel with the values of the parents.
        pp = new Panel();
        pp.setLayout(new GridLayout(parents_size - 1, 2));
        parents_labels = new Label[parents_size - 1];
        parent_choices = new Choice[parents_size - 1];

        // Fill the labels and choices for the values of the parents.
        for (int i = 0; i<parents_labels.length; i++) {
            // Get a parent
            parent_node = (InferenceGraphNode)(node_parents.elementAt(i+1));
            parents_labels[i] = new Label(parent_node.get_name());
            parents_labels[i].setForeground(color_parents);
            parent_choices[i] = new Choice();
            parent_choices[i].setForeground(color_parents);
            parent_node_values = parent_node.get_values();
            // Fill the choice item with the values for the parent.
            for (int j=0; j<parent_node_values.length; j++) {
                parent_choices[i].addItem(parent_node_values[j]);
            }
            // Insert the label and choice
            pp.add(parents_labels[i]);
            pp.add(parent_choices[i]);
        }

        // Just use a panel to put the title in the parent choices.
        ap = new Panel();
        ap.setLayout(new BorderLayout());
        parents_announce = new Label(parents_announce_label, Label.CENTER);
        parents_announce.setForeground(color_parents);
        ap.add("North", parents_announce);
        ap.add("Center", pp);
        separator_label = new Label("");
        ap.add("South", separator_label);

        // Now set the distribution panel
        build_distribution_panel_for_one_or_more_parents(node_values,
            first_parent_name, first_parent_values);

        // Finally set the panel with all choices and fields.
        gp = new Panel();
        gp.setLayout(new BorderLayout());
        gp.add("North", ap);
        gp.add("Center", dp);

        generate_credal_set_panel();

        // Fill the distribution
        fill_distribution();
    }

    /* ******************************************************* *
     * Build the widget when the variable has one parent.      *
     * ******************************************************* */
    private void build_with_one_parent() {
        Vector node_parents = node.get_parents();
      	String node_values[] = node.get_values();

        // Information about the first parent of the given node.
        InferenceGraphNode first_parent_node =
            (InferenceGraphNode)(node_parents.firstElement());
        String first_parent_name = first_parent_node.get_name();
        String first_parent_values[] = first_parent_node.get_values();

        // For a node one parent, do not
        // set the parent panel; go directly to the distribution panel.
        build_distribution_panel_for_one_or_more_parents(node_values,
            first_parent_name, first_parent_values);

        // Finally set the panel with all choices and fields.
        gp = new Panel();
        gp.setLayout(new BorderLayout());
        gp.add("Center", dp);

        generate_credal_set_panel();

        // Fill the distribution
        fill_distribution();
    }

    /* ******************************************************** *
     * Build the widget when the variable has no parents.       *
     * ******************************************************** */
    private void build_with_no_parent() {
      	String node_values[] = node.get_values();

        // For a node with no parent, do not
        // set the parent panel; go directly to the distribution panel.
        dp = new Panel();
        dp.setLayout(new GridLayout(node_values.length, 2));

        node_values_labels = new Label[node_values.length];
        fields = new TextField[node_values.length][1];

        for (int j=0; j<node_values.length; j++) {
            node_values_labels[j] = new Label(node_values[j]);
            dp.add(node_values_labels[j]);
            fields[j][0] = new TextField();
            dp.add(fields[j][0]);
        }

        // Set the panel with all choices and fields.
        gp = new Panel();
        gp.setLayout(new BorderLayout());
        gp.add("Center", dp);

        generate_credal_set_panel();

        // Fill the distribution
        fill_distribution();
    }

    /* ******************************************************* *
     * Build the distribution panel.                           *
     * ******************************************************* */
    private void build_distribution_panel_for_one_or_more_parents(String node_values[],
    String first_parent_name, String first_parent_values[]) {
        int i, j;

        dp = new Panel();
        dp.setLayout(new GridLayout(node_values.length + 1,
                                    first_parent_values.length + 1));
        node_values_labels = new Label[node_values.length];
        first_parent_values_labels = new Label[first_parent_values.length];
        fields = new TextField[node_values.length][first_parent_values.length];

        first_parent_name_label = new Label(first_parent_name);
        first_parent_name_label.setForeground(color_parents);
        dp.add(first_parent_name_label);
        for (i=0; i<first_parent_values.length; i++) {
            first_parent_values_labels[i] = new Label(first_parent_values[i]);
            first_parent_values_labels[i].setForeground(color_parents);
            dp.add(first_parent_values_labels[i]);
        }

        for (j=0; j<node_values.length; j++) {
            node_values_labels[j] = new Label(node_values[j]);
            dp.add(node_values_labels[j]);

            for (i=0; i<first_parent_values.length; i++) {
                fields[j][i] = new TextField();
                dp.add(fields[j][i]);
            }
        }
    }

    /* ****************************************************** *
     * Fill all the TextField objects with the values of the  *
     * correct distribution.                                  *
     * ****************************************************** */
    private void fill_distribution() {
     	int parents_size = node.get_parents().size();

     	if (parents_size > 1)
     	    fill_distribution_with_more_than_one_parent();
     	else if (parents_size == 1)
     	    fill_distribution_with_one_parent();
     	else if (parents_size == 0)
     	    fill_distribution_with_no_parent();
    }

    /* ******************************************************* *
     * Generate a panel for credal sets, if necessary.         *
     * ******************************************************* */
    private void generate_credal_set_panel() {
        if (node.is_credal_set()) {
            csp = new Panel();
            csp.setLayout(new BorderLayout());
            credal_set_label = new Label(credal_set);
            credal_set_choice = new Choice();
            for (int i=0; i<node.number_extreme_distributions(); i++)
                credal_set_choice.addItem( String.valueOf(i) );
            csp.add("West", credal_set_label);
            csp.add("Center", credal_set_choice);

            gp.add("South", csp);
        }
    }

    /* ****************************************************** *
     * Fill all the TextField when the variable has more than *
     * one parent.                                            *
     * ****************************************************** */
    private void fill_distribution_with_more_than_one_parent() {
        int i, j, k;
        String node_name = node.get_name();
        String node_values[] = node.get_values();
    	String variable_value_pairs[][] = new String[node.get_parents().size() + 1][2];

        if (node.is_credal_set())
            index_extreme_point = credal_set_choice.getSelectedIndex();

  	    for (i=0; i<parent_choices.length; i++) {
   	        variable_value_pairs[i][0] = parents_labels[i].getText();
   	        variable_value_pairs[i][1] = parent_choices[i].getSelectedItem();
   	    }

        variable_value_pairs[i][0] = first_parent_name_label.getText();
        variable_value_pairs[i+1][0] = node_name;

       	for (j=0; j<fields.length; j++) {
       	    for (k=0; k<fields[j].length; k++) {
                variable_value_pairs[i][1] = first_parent_values_labels[k].getText();
                variable_value_pairs[i+1][1] = node_values[j];
                fields[j][k].setText(
                    String.valueOf(node.get_function_value(variable_value_pairs,
                                                           index_extreme_point)) );
            }
        }
    }

    /* ****************************************************** *
     * Fill all the TextField when the variable has one       *
     * parent.                                                *
     * ****************************************************** */
    private void fill_distribution_with_one_parent() {
        int j, k;
        String node_name = node.get_name();
        String node_values[] = node.get_values();
    	String variable_value_pairs[][] = new String[2][2];

        if (node.is_credal_set())
            index_extreme_point = credal_set_choice.getSelectedIndex();

        variable_value_pairs[0][0] = first_parent_name_label.getText();
        variable_value_pairs[1][0] = node_name;

       	for (j=0; j<fields.length; j++) {
       	    for (k=0; k<fields[j].length; k++) {
                variable_value_pairs[0][1] = first_parent_values_labels[k].getText();
                variable_value_pairs[1][1] = node_values[j];
                fields[j][k].setText(
                    String.valueOf(node.get_function_value(variable_value_pairs,
                                                           index_extreme_point)) );
            }
        }
    }

    /* ****************************************************** *
     * Fill all the TextField when the variable has no        *
     * parent.                                                *
     * ****************************************************** */
    private void fill_distribution_with_no_parent() {
        int j;
        String node_name = node.get_name();
        String node_values[] = node.get_values();
    	String variable_value_pairs[][] = new String[1][2];

        if (node.is_credal_set())
            index_extreme_point = credal_set_choice.getSelectedIndex();

        variable_value_pairs[0][0] = node_name;

       	for (j=0; j<fields.length; j++) {
            variable_value_pairs[0][1] = node_values[j];
            fields[j][0].setText(
                String.valueOf(node.get_function_value(variable_value_pairs,
                                                       index_extreme_point)) );
        }
    }

    /* ************************************************** *
     * Handle the events.                                 *
     * ************************************************** */
    public boolean action(Event evt, Object arg) {
        int i, j, k;
        double value_set;
        String variable_value_pairs[][];
        String node_name = node.get_name();
        String node_values[] = node.get_values();
        int parents_size = node.get_parents().size();

        // Check whether to dismiss
        if (evt.target == return_button) {
	        dispose();
	        return(true);
	    }

        // In case the node has more than one
        // parent, check whether parent values
        // have been changed.
        if (parents_size > 1) {
            for (i=0; i<parent_choices.length; i++) {
	            if (evt.target == parent_choices[i]) {
	                fill_distribution();
	                return(true);
    	        }
            }
        }

        // Check whether the distribution has
        // been changed.
        for (i=0; i<fields.length; i++) {
	        for (j=0; j<fields[i].length; j++) {
	            if (evt.target == fields[i][j]) {
	                try {
                        value_set = (new Double( fields[i][j].getText() ).doubleValue() );
                    } catch (NumberFormatException e) {
                        return(true);
                    }

	                if (parents_size > 1) {
                    	variable_value_pairs =
                    	    new String[node.get_parents().size() + 1][2];
                  	    for (k=0; k<parent_choices.length; k++) {
                   	        variable_value_pairs[k][0] =
                   	            parents_labels[k].getText();
                   	        variable_value_pairs[k][1] =
                   	            parent_choices[k].getSelectedItem();
                   	    }
                        variable_value_pairs[k][0] =
                            first_parent_name_label.getText();
                        variable_value_pairs[k][1] =
                            first_parent_values_labels[j].getText();
                        variable_value_pairs[k+1][0] = node_name;
                        variable_value_pairs[k+1][1] = node_values[i];
                        node.set_function_value(variable_value_pairs, value_set,
                                                index_extreme_point);
                    }
                    else if (parents_size == 1) {
                    	variable_value_pairs = new String[2][2];
                        variable_value_pairs[0][0] =
                            first_parent_name_label.getText();
                        variable_value_pairs[0][1] =
                            first_parent_values_labels[j].getText();
                        variable_value_pairs[1][0] = node_name;
                        variable_value_pairs[1][1] = node_values[i];
                        node.set_function_value(variable_value_pairs, value_set,
                                                index_extreme_point);
                    }
	                else if (parents_size == 0) {
                    	variable_value_pairs = new String[1][2];
                        variable_value_pairs[0][0] = node_name;
                        variable_value_pairs[0][1] = node_values[i];
                        node.set_function_value(variable_value_pairs, value_set,
                                                index_extreme_point);
                    }

                    return(true);
                }
            }
        }

        // Check whether the index of the credal set has changed.
        if (evt.target == credal_set_choice) {
            index_extreme_point = credal_set_choice.getSelectedIndex();
            fill_distribution();
            return(true);
        }

        return(super.action(evt, arg));
    }
}

