/**
 *  EditNetworkDialog.java
 *
 *  Original version by  Sreekanth Nagarajan, rewritten
 *  by Fabio Cozman.
 *
 */

package JavaBayesInterface;

import InferenceGraphs.*;

import java.awt.*;
import java.util.*;

class EditNetworkDialog extends Dialog {
    InferenceGraph ig;

    int displayed_network_property_index;
    Panel np, npp, gnp, gncp, gnpp, tp, okp;
    Label name, network_properties;
    TextField text_name, text_global_parameter;
    Label global, global_parameter;
    CheckboxGroup globals;
    Checkbox no_global, epsilon_global, ratio_global, total_global, bounded_global;
    Button new_network_property, next_network_property;
    TextField network_properties_text;
    Button return_button;
    
    private final static int TOP_INSET = 5;
    private final static int LEFT_INSET = 10;
    private final static int RIGHT_INSET = 10;
    private final static int BOTTOM_INSET = 0;

    private final static String name_label = "Name:";
    private final static String network_properties_label = "Network properties:";
    private final static String next_property_label = "Next";
    private final static String new_property_label = "New";
    private final static String global_label = "Network neighborhood model:";
    private final static String no_global_label = "No global neighborhood";
    private final static String epsilon_global_label = "Epsilon contaminated neighborhood";
    private final static String ratio_global_label = "Constant density ratio neighborhood";
    private final static String total_global_label = "Total variation neighborhood";
    private final static String bounded_global_label = "Constant density bounded neighborhood";
    private final static String global_parameter_label = "Global neighborhood parameter:";
    private final static String return_label = "Close window";

    /* ****************************************************** *
     * Default constructor for an EditNetworkDialog object.  *
     * ****************************************************** */
    public EditNetworkDialog(Frame parent, InferenceGraph i_g) {
    	super(parent, "Edit: " + i_g.get_name(),
              false); // This used to be true, but bug in JDK1.1.6 does not accept it.
    	this.ig = i_g;

    	// Compose the frame

        // Panel for the name
        np = new Panel();
        np.setLayout(new BorderLayout());
        name = new Label(name_label);
        text_name = new TextField(30);
        np.add("West", name);
        np.add("Center", text_name);

        // Network properties
        npp = new Panel();
        npp.setLayout(new BorderLayout());
        network_properties = new Label(network_properties_label);
        next_network_property = new Button(next_property_label);
        new_network_property = new Button(new_property_label);
        network_properties_text = new TextField(40);

        npp.add("North", network_properties);
        npp.add("West", next_network_property);
        npp.add("Center", network_properties_text);
        npp.add("East", new_network_property);

        // Global neighborhood parameters
        gnp = new Panel();
        gnp.setLayout(new BorderLayout());
        global = new Label(global_label);

        gncp = new Panel();
        gncp.setLayout(new GridLayout(5, 1));
        globals = new CheckboxGroup();
        no_global = new Checkbox(no_global_label, globals, true);
        epsilon_global = new Checkbox(epsilon_global_label, globals, false);
        ratio_global = new Checkbox(ratio_global_label, globals, false);
        total_global = new Checkbox(total_global_label, globals, false);
        bounded_global = new Checkbox(bounded_global_label, globals, false);
        gncp.add(no_global);
        gncp.add(epsilon_global);
        gncp.add(ratio_global);
        gncp.add(total_global);
        gncp.add(bounded_global);

        gnpp = new Panel();
        gnpp.setLayout(new BorderLayout());
        global_parameter = new Label(global_parameter_label);
        text_global_parameter = new TextField(10);
        gnpp.add("West", global_parameter);
        gnpp.add("Center", text_global_parameter);

        gnp.add("North", global);
        gnp.add("Center", gncp);
        gnp.add("South", gnpp);

        // All the network parameters
        tp = new Panel();
        tp.setLayout(new BorderLayout());
    	tp.add("North", np);
    	tp.add("Center", npp);
    	tp.add("South", gnp);

        // Return buttons
     	okp = new Panel();
	    okp.setLayout(new FlowLayout(FlowLayout.CENTER));
        return_button =	new Button(return_label);
        okp.add(return_button);

        setLayout(new BorderLayout());
        add("North", tp);
        add("Center", okp);

	    // Pack the whole window
    	pack();

    	// Initialize values
    	fill_dialog();
    }

    /* ************************************************ *
     * Customized show method.                          *
     * ************************************************ */
    public synchronized void show() {
    	Rectangle bounds = getParent().bounds();
    	Rectangle abounds = bounds();

    	move(bounds.x + (bounds.width - abounds.width)/ 2,
    	     bounds.y + (bounds.height - abounds.height)/2);

    	super.show();
    }

    /* ************************************************** *
     * Customize insets method.                           *
     * ************************************************** */
    public Insets insets() {
        Insets ins = super.insets();
        return(new Insets(ins.top + TOP_INSET, ins.left + LEFT_INSET,
                          ins.bottom + BOTTOM_INSET, ins.right + RIGHT_INSET));
    }

    /* ************************************************** *
     * Fill the values in the dialog area.                *
     * ************************************************** */
    private void fill_dialog() {
        String values[], all_values = "";
        Vector prop;
        String property;
        double par;

        // Synchronize the network if necessary.
        ig.get_bayes_net();

        // Fill name
        text_name.setText(ig.get_name());

        // Fill network properties
        prop = ig.get_network_properties();
        if ((prop != null) && (prop.size() > 0)) {
            property = (String)(prop.firstElement());
            network_properties_text.setText(property);
            displayed_network_property_index = 0;
        }
        else
            displayed_network_property_index = -1;

        // Set global neighborhood
        switch (ig.get_global_neighborhood_type()) {
            case InferenceGraph.NO_CREDAL_SET:
            globals.setCurrent(no_global);
            break;
            case InferenceGraph.CONSTANT_DENSITY_RATIO:
            globals.setCurrent(ratio_global);
            break;
            case InferenceGraph.EPSILON_CONTAMINATED:
            globals.setCurrent(epsilon_global);
            break;
            case InferenceGraph.CONSTANT_DENSITY_BOUNDED:
            globals.setCurrent(bounded_global);
            break;
            case InferenceGraph.TOTAL_VARIATION:
            globals.setCurrent(total_global);
            break;
        }

        par = ig.get_global_neighborhood_parameter();
        text_global_parameter.setText( String.valueOf(par) );
    }

    /* ************************************************** *
     * Handle the observation events.                     *
     * ************************************************** */
    public boolean action(Event evt, Object arg) {
        Vector prop;
        String values[], property, checked_name;
        double par;

        if (evt.target == return_button)
	        dispose();
	    else if (evt.target == text_name) {
	        checked_name = ig.check_name(text_name.getText());
	        if (checked_name != null)
    	        ig.set_name(checked_name);
	    }
        else if (evt.target == no_global) {
            ig.set_global_neighborhood(InferenceGraph.NO_CREDAL_SET);
        }
        else if (evt.target == epsilon_global) {
            ig.set_global_neighborhood(InferenceGraph.EPSILON_CONTAMINATED);
        }
        else if (evt.target == ratio_global) {
            ig.set_global_neighborhood(InferenceGraph.CONSTANT_DENSITY_RATIO);
        }
        else if (evt.target == bounded_global) {
            ig.set_global_neighborhood(InferenceGraph.CONSTANT_DENSITY_BOUNDED);
        }
        else if (evt.target == total_global) {
            ig.set_global_neighborhood(InferenceGraph.TOTAL_VARIATION);
        }
        else if (evt.target == new_network_property) {
            displayed_network_property_index = -1;
            network_properties_text.setText("");
        }
        else if (evt.target == next_network_property) {
            prop = ig.get_network_properties();
            if ((prop != null) && (prop.size() > 0)) {
                displayed_network_property_index++;
                if (displayed_network_property_index >= prop.size())
                    displayed_network_property_index = 0;
                property = (String)(prop.elementAt(displayed_network_property_index));
                network_properties_text.setText(property);
            }
        }
        else if (evt.target == network_properties_text) {
            if (displayed_network_property_index != -1)
                ig.remove_network_property(displayed_network_property_index);
            ig.add_network_property(network_properties_text.getText());
        }
        else if (evt.target == text_global_parameter) {
            try {
                par = (new Double( text_global_parameter.getText() ).doubleValue() );
            } catch (NumberFormatException e) {
                return(true);
            }
            if (par <= 0.0) par = 0.0;
            ig.set_global_neighborhood_parameter(par);
        }
        else return super.action(evt, arg);

    	return true;
    }

}

