package learn;

import java.util.Enumeration;
import java.util.Hashtable;

/**
 * This will implement Q-learning
 * <p>Title: Distributed AI practicum</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Coen Pieterse
 * @version 1.0
 */



public class QValueTable {

	/**
	 * if no value was found , because the state is not found
	 */
	public static final int ACTION_NOT_FOUND = Integer.MIN_VALUE; 

	/**
	 * if no value was found , because the state is not found
	 */
	public static final double STATE_NOT_FOUND = -Double.MAX_VALUE; 

	/**
	 * Take a random action
	 */
	public static final int ACTION_RANDOM = Integer.MAX_VALUE; 
	
  /**
   * labda will set the influence of new data
   * this can change during the agent's life, this is
   * specified by the counter
   * @param counter parameter to the labda function
   * @return labda
   */
  private double labda( double counter ){
    //return Math.exp( - counter / 5 );
    return 0.5;
  }

  /**
   * Gamma weights the influence of the estimation of the next state
   * @param counter parameter to the gamma function
   * @return gamma
   */
  private double gamma( double counter ){
   return 0.85;
  }

  /**
   * epsilon models the chance that the agent explores a different action
   * this chance is decreased during life by the counter
   * @param counter parameter to the epsilon function
   * @return epsilon
   */
  private double epsilon( double counter ){
    return Math.exp( - 3 * counter );
	//  return 0.1;
  }

  /**
   * this hashtable maps a state to the current estimated utlity of the state
   * for each action. the utilities are given in an Vector array, in which each
   * entry corresponds to the given Q value for an action, given the state
   * The array entries are
   * 0 --> move none
   * 1 --> move up
   * 2 --> move east
   * 3 --> move down
   * 4 --> move west
   */
  private java.util.Hashtable stateQTable = new java.util.Hashtable( 50 );


  /**
   * main constructor, builds the first piece of the hashtable
   */
  public QValueTable( ){
  }

  /**
   * add a new entry to the table for the given state
   * @param state the state for which the entry is added
   * @param reward set the Q value to the immediate reward
   */
  private void addToTable( QState state ){
    //System.out.println("State added");
    TableEntry entry = new TableEntry();
    stateQTable.put(state, entry);
  }

  /**
   * get current evaluation of the state-action pair
   * @param state current state the agent is in
   * @param action the action it wants to take
   * @return evaluation
   */
  public double getQ( QState state, int action ){
    // check if the state is already in the table
    if( !stateQTable.containsKey(state) )
      addToTable(state);
    // retreive current assignments
    TableEntry entry = (TableEntry) stateQTable.get(state);
    return entry.getQ( action );
  }

  /**
   * get current evaluation of the maximum action for this state
   * @param state state the agent is in
   * @return current maximum evaluation
   */
  public double getMaxQ( QState state ){
    // check if the state is already in the table
    //System.out.println(" check if state is in the table ");
    if( !stateQTable.containsKey(state) )
    {
	  addToTable( state );
      return STATE_NOT_FOUND;
    }
    // retreive current assignments
    TableEntry entry = (TableEntry) stateQTable.get(state);

    // initialize maximum
    Enumeration enumerate = entry.getActions();
    
    double value = STATE_NOT_FOUND;
    double max = -Double.MAX_VALUE;
    double tmpValue;
    // loop through vector
    while( enumerate.hasMoreElements() )
    {
	  Integer keyAction = (Integer)enumerate.nextElement();
      // check if value > max
	  tmpValue = entry.getQ( keyAction ); 
      if ( tmpValue > max ){
        // assign new maximum
        max = tmpValue;
        value = tmpValue;
      }    	  
    } // end for
    // return found maximum
    return value;
  }

  /**
   * get current evaluation of the maximum action for this state
   * @param state state the agent is in
   * @return current maximum evaluation
   */
  public int getMaxAction( QState state ){
    // check if the state is already in the table
    if( !stateQTable.containsKey(state) )
    {
      addToTable(state);
      return ACTION_RANDOM;
    }
    //System.out.println("Requesting max action");
      // retreive current assignments
    TableEntry entry = (TableEntry) stateQTable.get(state);

    //System.out.println( "Epsilon:  " + epsilon(entry.getRelGlobalCount()) );
    // sometimes explore new actions epsilon, estimated with
    // respect to the relative count of the first action
    if( Math.random() < epsilon(entry.getRelGlobalCount()) )
      return ACTION_RANDOM;

    // initialize maximum
    Enumeration enumerate = entry.getActions();
    
    int value = ACTION_RANDOM;
    double max = -Double.MAX_VALUE;
    double tmpValue;
    // loop through vector
    while( enumerate.hasMoreElements() )
    {
	  Integer keyAction = (Integer)enumerate.nextElement();
      // check if value > max
	  tmpValue = entry.getQ( keyAction ); 
      if( tmpValue > max ){
    	// store int value
    	value = keyAction.intValue();
    	max = tmpValue;
      }
      else if( tmpValue == max && Math.random() > 0.5 ){
      	// store int value
      	value = keyAction.intValue();
      }
    }
    // return found maximum
    return value;
  }

  /**
   * Adjust evaluation of a state-action pair, to fit the new data
   * @param state state the agent was in
   * @param action the action the agent took
   * @param R reward of the resulting state
   * @param Q evaluation of the maximum action of the resulting state
   */
  public void update( QState state, int action, double R, double Q ){
	if( Q == STATE_NOT_FOUND )
	  Q = 0;
    // retreive current assignments
    TableEntry entry = (TableEntry) stateQTable.get(state);
    // get relative count for this action
    double relativeCount = entry.getRelCount(action);
    // get labda, take as paramter the relative count for this action
    double labda = labda( relativeCount );
    // evaluate new value
    double newQ = ( 1 - labda ) * entry.getQ(action)
        + labda * ( R + gamma( relativeCount ) * Q );
    // update value
    entry.setQ( action, newQ );
  }

  /**
   * Adjust evaluation of a state-action pair, to fit the new data
   * @param state state the agent was in
   * @param action the action the agent took
   * @param R reward of the resulting state
   * @param Q evaluation of the maximum action of the resulting state
   */
  public void update( QState state, int action, double R, QState resultState ){
    // retreive current assignments
    TableEntry entry = (TableEntry) stateQTable.get(state);
	// get q value if we take the best possible action
	double Q = getMaxQ( resultState );
	// new sate
	if( Q == STATE_NOT_FOUND )
	{
		Q = 0;
	}
		
    // get relative count for this action
    double relativeCount = entry.getRelCount(action);
    // get labda, take as paramter the relative count for this action
    double labda = labda( relativeCount );
    // evaluate new value
    double newQ = ( 1 - labda ) * entry.getQ(action)
        + labda * ( R + gamma( relativeCount ) * Q );
    
    // update value
    entry.setQ( action, newQ );
  }


  /**
   * Each instance represents an entry in the table
   * containing an double array of q values for each move
   * and a counter
   * <p>Title: Distributed AI practicum</p>
   * <p>Description: </p>
   * <p>Copyright: Copyright (c) 2003</p>
   * <p>Company: </p>
   * @author Coen Pieterse
   * @version 1.0
   */
  private class TableEntry{

    /**
     * set the maximum for the counters
     */
    private final int MAXCOUNTER = 50;

    /**
     * counter for amount of calls to this state
     */
    private int globalCounter = 0;
    /**
     * counter to count how many times the value has been updated
     */
    private Hashtable counter = new Hashtable();

    /**
     * The doubles represent the Q function value for the action
     */
    private Hashtable q = new Hashtable();

    /**
     * get current estimation of the q value for this action
     * @param action the action to evaluate
     * @return current estimtation
     */
    public double getQ( int action ){
      Double d = (Double)q.get( new Integer( action ) );
      if( d == null )
    	  return 0;
      else
    	  return d.doubleValue();
    }

    /**
     * get current estimation of the q value for this action
     * @param action the action to evaluate
     * @return current estimtation
     */
    public double getQ( Integer action ){
      Double d = (Double)q.get( action );
      if( d == null )
    	  return 0;
      else
    	  return d.doubleValue();
    }

    /**
     * Get the available actions
     * @return
     */
    public Enumeration getActions()
    {
    	return q.keys();
    }

    /**
     * set new estimation for the action,
     * increments counter as well
     * @param action the action to update
     * @param qValue new estimation
     */
    public void setQ( int action, double qValue ){
      Integer key = new Integer(action);
      Integer count = (Integer)counter.get(key);
      
      if( count == null )
    	  count = new Integer(0);
      
      int c = count.intValue();
      
      if( c < MAXCOUNTER ){
        c++;
        globalCounter++;
        counter.put( key, new Integer(c) );
        q.put( key, new Double(qValue) );
      }
    }

    /**
     * get current counter for this action
     * @param action the action
     * @return counter for this action
     */
    public int getCount( int action ){
        Integer i = (Integer)counter.get( new Integer( action ) );
        if( i == null )
      	  return 0;
        else
      	  return i.intValue();
    }

    /**
     * get current counter for this action
     * divided by the maximum count
     * @param action the action
     * @return counter for this action
     */
    public double getRelCount( int action ){
      return getCount(action)*1.0/MAXCOUNTER;
    }

    /**
     * get current global counter for this action
     * divided by the maximum count
     * @return counter for this action
     */
    public double getRelGlobalCount( ){
      return (double)globalCounter*2.0/(counter.size()*MAXCOUNTER);
    }


  }

}